"""Define Abstract class for Model with neccessary methods and methods to implement."""
from abc import ABC, abstractmethod

from ML_management import mlmanagement
from ML_management.models.action_type import ActionType

import mlflow


class Model(mlflow.pyfunc.PythonModel, ABC):
    """Abstract class for model that Job will use."""

    def __init__(self):
        """Get model from MLflow or local."""
        self._test_data_path = None

    @property
    def test_data_path(self):
        """Data path property getter."""
        # do i need to raise if self._data_path is uninitialized?
        return self._test_data_path

    @test_data_path.setter
    def test_data_path(self, path: str):
        """Data path property setter."""
        self._test_data_path = path

    @abstractmethod
    def predict_function(self):
        """Every model should make predictions."""
        raise NotImplementedError

    def upload_model(
        self,
        pyfunc_model_name: str,  # TODO weird argument, get rid of it
        artifacts: dict = None,
        registered_model_name: str = None,
    ):
        """Upload wrapper to MLmanagement server."""
        with mlmanagement.start_run(nested=True):
            mlmanagement.log_model(
                artifact_path=pyfunc_model_name,
                python_model=self,
                artifacts=artifacts,
                registered_model_name=self.model_name if hasattr(self, "model_name") else registered_model_name,
                action_type=ActionType.UPLOAD,  # should this be the same as job type? can also put it into class. OR, make it new type.
                source_model_name=self.source_model_name
                if hasattr(self, "source_model_name")
                else None,  # set that after model download from mlflow
                source_model_version=self.source_model_version
                if hasattr(self, "source_model_version")
                else None,  # set that after model download from mlflow
            )

    def set_experiment_name(self, experiment_name: str):
        """Set experiment name and return experiment."""
        mlmanagement.set_experiment(experiment_name)
        experiment = mlmanagement.get_experiment_by_name(experiment_name)
        return experiment

    @abstractmethod
    def get_data_flavour(self):
        """Every model is required to store one of the data flavour."""
        raise NotImplementedError



"""Custom exception definition for server mlflow manager graph."""


class InvalidActionType(Exception):
    """Define InvalidActionType exception."""

    def __init__(self, passed_action_type):
        self.passed_action_type = passed_action_type
        self.message = (
            'Passed action_type "{passed_action_type}" is invalid,'
            "must be Enum value from ML_management.models.action_type.ActionType".format(passed_action_type=passed_action_type)
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidActionType, (self.passed_action_type,))


class InvalidModelName(Exception):
    """Define InvalidModelName exception."""

    def __init__(self, passed_name, kwarg):
        self.passed_name = passed_name
        self.kwarg = kwarg
        self.message = 'Passed "{kwarg}" argument value "{passed_name}" is invalid, as this model is not registered in mlflow.'.format(
            kwarg=kwarg,
            passed_name=passed_name,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidModelName, (self.passed_name, self.kwarg))


class InvalidModelVersion(Exception):
    """Define InvalidModelVersion exception."""

    def __init__(self, passed_name, passed_version, name_kwarg, version_kwarg):
        self.passed_version = passed_version
        self.passed_name = passed_name
        self.name_kwarg = name_kwarg
        self.version_kwarg = version_kwarg
        self.message = 'Passed "{version_kwarg}" argument value "{passed_version}" is invalid, \
            as model "{passed_name}" (passed as "{name_kwarg}" argument value) has no such version.'.format(
            passed_name=passed_name,
            passed_version=passed_version,
            name_kwarg=name_kwarg,
            version_kwarg=version_kwarg,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (
            InvalidModelVersion,
            (
                self.passed_name,
                self.passed_version,
                self.name_kwarg,
                self.version_kwarg,
            ),
        )


class KwargNotPassedWithAction(Exception):
    """Define KwargNotPassedWithAction exception."""

    def __init__(self, kwarg, passed_action_type):
        self.passed_action_type = passed_action_type
        self.kwarg = kwarg
        self.message = 'Argument "{kwarg}" cannot be ommitted with action type ActionType.{passed_action_type}.'.format(
            kwarg=kwarg,
            passed_action_type=passed_action_type,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (KwargNotPassedWithAction, (self.kwarg, self.passed_action_type))



"""List of action types for vertex of version graph."""
from enum import Enum


class ActionType(str, Enum):
    """List of action types for vertex of version graph."""

    UPLOAD = "upload"
    TRAIN = "train"
    FINETUNE = "finetune"



"""Dummy dataset inherited from the base abstract class."""
from ML_management.dataset.datamodel import DatasetFlavour
from ML_management.dataset.Dataset import Dataset


class DummyDataset(Dataset):
    """Implementation of DummyLoader Dataset."""

    def get_json_schema(self):
        """Return empty json schema."""
        schema = {}
        return schema

    def set_data(self, *args, **kwargs):
        """Return empty string."""
        return ""

    def get_data_flavour():
        """Return flavour."""
        return DatasetFlavour.DummyDataset
